//----------------------------------------------------------------------//
// MIDI|[gNX														//
//																		//
//----------------------------------------------------------------------//
#include <tchar.h>
#include "MidiPort.h"

//---------------------------------------------------------------------------
TMidiPort *MidiPort;

//---------------------------------------------------------------------------
//GMFXg
const char GM_Instrument[128][32] = {
	"Grand Pf","Bright Pf","E.Grand Pf","Honky-tonk",
	"E.Pf1","E.Pf2","Harpsichord","Clavi",
	"Celesta","Glocken","Music Box","Vibraphone",
	"Marimba","Xylophone","Tubular Bel","Dulcimer",
	"Draw. Organ","Perc. Organ","Rock Organ","Church Organ",
	"Reed Organ","Accordion","Harmonica","Tango Accordion",
	"Nylon Gtr","Steel Gtr","E.Gtr (Jazz)","E.Gtr (Clean)",
	"E.Gtr (Muted)","Overdriven Gtr","Dist. Gtr","Gtr Harmo.",
	"Aco. Bass","E.Bass(Fingar)","E.Bass(Pick)","Fretless Bass",
	"Slap Bass1","Slap Bass2","Syn. Bass1","Syn. Bass2",
	"Violin","Viola","Cello","Contrabass",
	"Trem. Strings","Pizz. Strings","Harp","Timpani",
	"String1","String2","Syn. Strings1","Syn. Strings2",
	"Choir Aahs","Voice Oohs","Syn. Voice","Orchestra Hit",
	"Trumpet","Trombone","Tuba","Muted Trumpet",
	"French Horn","Brass Section","Syn. Brass1","Syn. Brass2",
	"Soprano Sax","Alto Sax","Tenor Sax","Baritone Sax",
	"Oboe","English Horn","Bassoon","Clarinet",
	"Piccolo","Flute","Recorder","Pan Flute",
	"Blown Bottle","Shakuhach","Whistle","Ocarina",
	"Lead1 square","Lead2 Sawtooth","Lead3 Calliope","Lead4 Chiff",
	"Lead5 Charang","Lead6 Voice","Lead7 Fifths","Lead8 Bass+Ld",
	"Pad1 New age","Pad2 Warm","Pad3 PolySyn.","Pad4 Choir",
	"Pad5 Bowed","Pad6 Metallic","Pad7 Halo","Pad8 Sweep",
	"FX1 Rain","FX2 Soundtrack","FX3 Crystal","FX4 Atmosphere",
	"FX5 Brightness","FX6 Goblins","FX7 Echoes","FX8 Sci-Fi",
	"Sitar","Banjo","Shamisen","Koto",
	"Kalimba","Bag Pipe","Fiddle","Shanai",
	"Tinkle Bell","Agogo","Steel Drums","Woodblock",
	"Taiko Drum","Melodic Tom","Syn. Drum","Reverse Cymbal",
	"Fret Noise","Breath Noise","Seashore","Bird Tweet",
	"Telephone","Helicopter","Applause","Gunshot"
};

//---------------------------------------------------------------------------
TMidiPort::TMidiPort()
{
	hmo = NULL;
	DevNum = 0;

	for (int i=0;i<MAX_MIDICHN;i++)	{
		ChnPrg[i] = 0;
		ChnVol[i] = 90;
		ChnPan[i] = 64;
	}

	for (int i=0;i<MAX_TONE;i++) tone[i].flag = 0;

	Recording = Recorded = false;
	RecData = NULL;
}
//---------------------------------------------------------------------------
TMidiPort::~TMidiPort()
{
	if (Recording) RecStop();
	if (RecData) {
		//Cxgf[^j
		for (int i=0;i<RecData->Count;i++) {
			evt_rec *ep = (evt_rec*)RecData->Items[i];
			if (ep->ExData!=NULL) delete ep->ExData;
			delete ep;
		}
		delete RecData;
	}

	if (hmo) OutClose();
}

//---------------------------------------------------------------------------
//MIDI o̓foCX擾
int __fastcall TMidiPort::GetDevMax()
{
	return midiOutGetNumDevs();
}

//---------------------------------------------------------------------------
//MIDI o̓foCXI[v
MMRESULT __fastcall TMidiPort::OutOpen()
{
	MMRESULT ret = midiOutOpen(&hmo,DevNum,NULL,NULL,CALLBACK_NULL);
	if (ret != MMSYSERR_NOERROR) hmo = NULL;
	return ret;
}
//---------------------------------------------------------------------------
//MIDI o̓foCXN[Y
void __fastcall TMidiPort::OutClose()
{
	if (hmo==NULL) return;

	for (int i=0;i<MAX_MIDICHN;i++) AllNoteOff(i);
	midiOutClose(hmo);
	hmo = NULL;
}
//---------------------------------------------------------------------------
//bZ[Wo
UINT __fastcall TMidiPort::sendMsg(BYTE bStatus, BYTE bData1,  BYTE bData2)
{
	if (hmo==NULL) return MMSYSERR_INVALHANDLE;

	if (Recording) {
		//CxgL^
		evt_rec *ep = new evt_rec;
		ep->count  = RecCount;
		ep->Status = bStatus;
		ep->Data1  = bData1;
		ep->Data2  = bData2;
		ep->ExData = NULL;
		RecData->Add(ep);
	}

	union {
		DWORD dwData;
		BYTE bData[4];
	} u;

    // Construct the MIDI message.
	u.bData[0] = bStatus;  // MIDI status byte
	u.bData[1] = bData1;   // first MIDI data byte
	u.bData[2] = bData2;   // second MIDI data byte
	u.bData[3] = 0;

	// Send the message. 
	return midiOutShortMsg(hmo, u.dwData); 
}

//---------------------------------------------------------------------------
//F擾
int __fastcall TMidiPort::GetPrg(int chn)
{
	return ChnPrg[chn];
}
//---------------------------------------------------------------------------
//Fݒ
void __fastcall TMidiPort::SetPrg(int chn, int prg)
{
	sendMsg((BYTE)(0xc0+chn), (BYTE)prg, 0);
	ChnPrg[chn] = prg;
}

//---------------------------------------------------------------------------
//
void __fastcall TMidiPort::SetVolume(int chn, int vol)
{
	sendMsg((BYTE)(0xb0+chn),7,(BYTE)(vol & 0x7f));
	ChnVol[chn] = vol;
}
//---------------------------------------------------------------------------
//p
void __fastcall TMidiPort::SetPan(int chn, int pan)
{
	sendMsg((BYTE)(0xb0+chn),10,(BYTE)(pan & 0x7f));
	ChnPan[chn] = pan;
}

//---------------------------------------------------------------------------
//m[gEI
void __fastcall TMidiPort::KeyOn(int chn, int key, int vel)
{
	sendMsg((BYTE)(0x90+chn),(BYTE)key,(BYTE)vel);
}
//---------------------------------------------------------------------------
//I[Em[gEIt
void __fastcall TMidiPort::AllNoteOff(int chn)
{
	sendMsg((BYTE)(0xb0+chn),123,0);
}

//---------------------------------------------------------------------------
// sb`xhZVreB[̐ݒ
void __fastcall TMidiPort::BendSensibility(int chn, int v)
{
	sendMsg((BYTE)(0xb0+chn),101,  0);		//RPN MSB
	sendMsg((BYTE)(0xb0+chn),100,  0);		//RPN LSB
	sendMsg((BYTE)(0xb0+chn),  6,(BYTE)v);	//DATA ENTRY MSB
	sendMsg((BYTE)(0xb0+chn), 36,  0);		//DATA ENTRY LSB
	sendMsg((BYTE)(0xb0+chn),127,127);
}
//---------------------------------------------------------------------------
//sb`xh (e)
void __fastcall TMidiPort::PitchBend(int chn, int bnd)
{
	sendMsg((BYTE)(0xe0+chn),0,(BYTE)bnd);
}
//---------------------------------------------------------------------------
//sb`xh ()
void __fastcall TMidiPort::PitchBendL(int chn, int bnd)
{
	sendMsg((BYTE)(0xe0+chn),(BYTE)(bnd & 0x7f),(BYTE)((bnd>>7) & 0x7f));
}

//---------------------------------------------------------------------------
//̐
//---------------------------------------------------------------------------
int  __fastcall TMidiPort::cre_ton(int c, int k, int v,int t, int o)
{
	int tn = -1;
	for (int i=0; (i<MAX_TONE) && (tn==-1); i++) {
		if (tone[i].flag==0) tn = i;
	}
	if (tn!=-1) {
		tone[tn].chn  = c;
		tone[tn].key  = k;
		tone[tn].vel  = v;
		tone[tn].tcnt = t;
		tone[tn].ownr = o;
		tone[tn].flag = 1;
		sendMsg((BYTE)(0x90+c), (BYTE)k, (BYTE)v);
		return tn;
	}
	else
		return -1;
}

//---------------------------------------------------------------------------
//̍XV
void __fastcall TMidiPort::upd_ton()
{
	for (int i=0;i<MAX_TONE;i++) {
		if ((tone[i].flag>0) && (tone[i].tcnt>0)) {
			if ((--(tone[i].tcnt))==0) {
				sendMsg((BYTE)(0x90+tone[i].chn), (BYTE)(tone[i].key), 0);
				tone[i].flag = 0;
			}
		}
	}
}

//---------------------------------------------------------------------------
//̉SĎ~߂
void __fastcall TMidiPort::stp_ton()
{
	for (int i=0;i<MAX_TONE;i++) {
		if (tone[i].flag==0) continue;
		sendMsg((BYTE)(0x90+tone[i].chn), (BYTE)(tone[i].key), 0);
		tone[i].tcnt = 0;
		tone[i].flag = 0;
	}
}


//---------------------------------------------------------------------------
//wI[i[̔擾
int __fastcall TMidiPort::get_tns(int o)
{
	int n = 0;
	for (int i=0;i<MAX_TONE;i++) {
		if (tone[i].ownr != o) continue;
		if ((tone[i].flag>0) && (tone[i].tcnt>0)) n++;
	}
	return n;
}

//---------------------------------------------------------------------------
//sb`xh ()
void __fastcall TMidiPort::bnd_ton(int id,int bnd)
{
	if (tone[id].flag>0) {
		sendMsg((BYTE)(0xe0+tone[id].chn), (BYTE)((bnd>>7) & 0x7f), (BYTE)(bnd & 0x7f));
	}
}


//===========================================================================
// MIDIL^^ WMIDIt@C쐬
//===========================================================================
void __fastcall TMidiPort::UpdateRecCount()
{
	if (Recording) RecCount = (::timeGetTime() - StartCount)/25;
}

//---------------------------------------------------------------------------
// L^t@Cւ̃f[^
// v: l  b:oCg(1`4)
//---------------------------------------------------------------------------
void __fastcall TMidiPort::WriteRecValue(int v, int b)
{
	if ((b<1) || (b>4)) return;

	int tmp = v;
	BYTE vbuf[4];

	for (int i=0; i<b; i++) {
		vbuf[i] = (BYTE)(tmp & 0xff);
		tmp >>= 8;
	}

	for (int i=b-1; i>=0; i--) putc(vbuf[i], RecStream);
}

//---------------------------------------------------------------------------
// L^Jn
//---------------------------------------------------------------------------
void __fastcall TMidiPort::RecStart(int iflag)
{
	int i;
	if (RecData) {
		//Cxgf[^j
		for (i=0;i<RecData->Count;i++) {
			evt_rec *ep = (evt_rec*)RecData->Items[i];
			if (ep->ExData!=NULL) delete ep->ExData;
			delete ep;
		}
		RecData->Clear();
	}
	else
		RecData = new TList();

	::timeBeginPeriod(1);
	StartCount = ::timeGetTime();
	Recorded   = false;
	Recording  = true;

	//
	if (iflag!=0) {
		RecCount = 0;
		if ((iflag & RECINI_PRG) != 0)
			for (i=0;i<MAX_MIDICHN;i++) SetPrg(i, ChnPrg[i]);
		RecCount = 1;
		if ((iflag & RECINI_VOL) != 0)
			for (i=0;i<MAX_MIDICHN;i++)	SetVolume(i, ChnVol[i]);
		RecCount = 2;
		if ((iflag & RECINI_PAN) != 0)
			for (i=0;i<MAX_MIDICHN;i++)	SetPan(i, ChnPan[i]);

		while (RecCount<5) {
			::Sleep(50);
			UpdateRecCount();
		}
	}
}

//---------------------------------------------------------------------------
// L^~
//---------------------------------------------------------------------------
void __fastcall TMidiPort::RecStop()
{
	if (Recording) {
		Recording = false;
	 	::timeEndPeriod(1);
	}
	Recorded = (RecData!=NULL)? (RecData->Count>0) : false;
}

//---------------------------------------------------------------------------
// L^N[Y  t@C֏o
//---------------------------------------------------------------------------
bool __fastcall TMidiPort::RecSave(UnicodeString fname)
{
	if (Recording) RecStop();
	if (!Recorded) return false;
	if (fname.IsEmpty()) return false;

	//L^MIDIt@CJ
	RecStream = _tfopen(fname.c_str(), _T("wb"));
	if (RecStream) {
		//--------------------------------
		//wb_EubN̏
		//--------------------------------
		fputs("MThd", RecStream);
		WriteRecValue( 6, 4);	//ubN = 6
		WriteRecValue( 0, 2);	//tH[}bg0
		WriteRecValue( 1, 2);	//1gbN
		WriteRecValue(20, 2);	//20/l

		//--------------------------------
		//ubNƃf^ݒ
		//--------------------------------
		int TrackDataLen = 0;
		int LastStatus	 = 0;
		int LastCount	 = 0;

		for (int i=0;i<RecData->Count;i++) {
			evt_rec *ep = (evt_rec*)RecData->Items[i];
			//f[^oCg̒𔻒f
			if (ep->Status == 0xf0) {
				//sysex Cxg
				TrackDataLen += (ep->datan+1);
			}
			else {
				//MIDI Cxg
				switch (ep->Status & 0xf0) {
				case 0xc0:	ep->datan = 1; break;
				case 0xd0:	ep->datan = 1; break;
				default:	ep->datan = 2;
				}
				TrackDataLen += ep->datan;
				//jOEXe[^X̃`FbN
				if (LastStatus!=ep->Status) {
					TrackDataLen ++;
					ep->running = false;
				}
				else
					ep->running = true;
			}
			LastStatus = ep->Status;

			//f^E^C߂
			int v = ep->count - LastCount;
			if (v<0) v=0;	//?
			int tmp = v & 0x7f;
			while ((v>>=7) != 0) {
				tmp <<= 8;
				tmp |= 0x80;
				tmp += (v & 0x7f);
			}
			ep->deltan = 0;
			for (int dn=0;dn<4;dn++) {
				ep->delta[dn] = (tmp & 0xff);
				ep->deltan ++;
				if ((tmp & 0x80)==0) break;
				tmp >>= 8;
			}
			TrackDataLen += ep->deltan;

			LastCount = ep->count;
		}

		TrackDataLen += 8;	//^ECxg(qL)
		TrackDataLen += 7;	//^ECxg(e|)
		TrackDataLen += 4;	//^ECxg(gbN̏I)

		//--------------------------------
		//gbNEubN̏
		//--------------------------------
		fputs("MTrk", RecStream);
		WriteRecValue(TrackDataLen, 4);	//ubN

		//^ECxg: qL
		putc(0x00, RecStream);  //delta = 0
		putc(0xff, RecStream);
		putc(0x58, RecStream);
		putc(0x04, RecStream);
		putc(0x04, RecStream);	// 4/4
		putc(0x02, RecStream);
		putc(0x18, RecStream);
		putc(0x08, RecStream);
		//e|
		putc(0x00, RecStream);  //delta = 0
		putc(0xff, RecStream);
		putc(0x51, RecStream);
		putc(0x03, RecStream);
		putc(0x07, RecStream);	//500,000us
		putc(0xa1, RecStream);
		putc(0x20, RecStream);

		//Cxg
		for (int i=0;i<RecData->Count;i++) {
			evt_rec *ep = (evt_rec*)RecData->Items[i];
			//f^E^C
			for (int dn=0;dn<ep->deltan;dn++) WriteRecValue(ep->delta[dn],1);

			if (ep->Status==0xf0) {
				//sysex Cxg
				WriteRecValue(0xf0,	1);
				WriteRecValue(ep->datan-1,	1);
				for (int j=1; j<ep->datan; j++) {
					WriteRecValue(ep->ExData[j],	1);
				}
			}
			else {
				//MIDI Cxg
				if (!ep->running)	WriteRecValue(ep->Status,	1);
									WriteRecValue(ep->Data1,	1);
				if (ep->datan==2)	WriteRecValue(ep->Data2,	1);
			}
		}

		//^ECxg: gbN̏I
		putc(0x00, RecStream);
		putc(0xff, RecStream);
		putc(0x2f, RecStream);
		putc(0x00, RecStream);

		//L^MIDIt@C
		fclose(RecStream);
		Recorded = false;
		return true;

	}
	else
		return false;
}
//---------------------------------------------------------------------------
